@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2012 inria.  All rights reserved.
+ * Copyright © 2009-2013 Inria.  All rights reserved.
  * Copyright © 2009-2012 Université Bordeaux 1
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -38,25 +38,25 @@ hwloc_calc_append_cpuset(hwloc_bitmap_t set, hwloc_const_bitmap_t newset,
   hwloc_bitmap_asprintf(&s2, set);
   switch (mode) {
   case HWLOC_CALC_APPEND_ADD:
-    if (verbose)
+    if (verbose > 0)
       fprintf(stderr, "adding %s to %s\n",
           s1, s2);
     hwloc_bitmap_or(set, set, newset);
     break;
   case HWLOC_CALC_APPEND_CLR:
-    if (verbose)
+    if (verbose > 0)
       fprintf(stderr, "clearing %s from %s\n",
           s1, s2);
     hwloc_bitmap_andnot(set, set, newset);
     break;
   case HWLOC_CALC_APPEND_AND:
-    if (verbose)
+    if (verbose > 0)
       fprintf(stderr, "and'ing %s from %s\n",
           s1, s2);
     hwloc_bitmap_and(set, set, newset);
     break;
   case HWLOC_CALC_APPEND_XOR:
-    if (verbose)
+    if (verbose > 0)
       fprintf(stderr, "xor'ing %s from %s\n",
           s1, s2);
     hwloc_bitmap_xor(set, set, newset);
@@ -96,6 +96,7 @@ hwloc_obj_type_sscanf(const char *string, hwloc_obj_type_t *typep, int *depthatt
   hwloc_obj_type_t type = (hwloc_obj_type_t) -1;
   int depthattr = -1;
   hwloc_obj_cache_type_t cachetypeattr = (hwloc_obj_cache_type_t) -1; /* unspecified */
+  char *end;
 
   /* types without depthattr */
   if (!hwloc_strncasecmp(string, "system", 2)) {
@@ -122,17 +123,29 @@ hwloc_obj_type_sscanf(const char *string, hwloc_obj_type_t *typep, int *depthatt
     type = HWLOC_OBJ_CACHE;
   } else if ((string[0] == 'l' || string[0] == 'L') && string[1] >= '0' && string[1] <= '9') {
     type = HWLOC_OBJ_CACHE;
-    depthattr = atoi(string+1);
-    if (string[2] == 'd')
+    depthattr = strtol(string+1, &end, 10);
+    if (*end == 'd') {
       cachetypeattr = HWLOC_OBJ_CACHE_DATA;
-    else if (string[2] == 'i')
+      end++;
+    } else if (*end == 'i') {
       cachetypeattr = HWLOC_OBJ_CACHE_INSTRUCTION;
+      end++;
+    } else if (*end == 'u') {
+      cachetypeattr = HWLOC_OBJ_CACHE_UNIFIED;
+      end++;
+    }
+    if (*end && hwloc_strncasecmp(end, "cache", 2))
+      return -1;
+
   } else if (!hwloc_strncasecmp(string, "group", 1)) {
     int length;
     type = HWLOC_OBJ_GROUP;
     length = strcspn(string, "0123456789");
-    if (string[length] != '\0')
-      depthattr = atoi(string+length);
+    if (string[length] != '\0') {
+      depthattr = strtol(string+length, &end, 10);
+      if (*end)
+	return -1;
+    }
   } else
     return -1;
 
@@ -160,14 +173,16 @@ hwloc_calc_depth_of_type(hwloc_topology_t topology, hwloc_obj_type_t type,
       return -1;
     depth = hwloc_get_type_or_above_depth(topology, type);
     if (depth == HWLOC_TYPE_DEPTH_MULTIPLE) {
-      fprintf(stderr, "type %s has multiple possible depths\n", hwloc_obj_type_string(type));
+      if (verbose >= 0)
+	fprintf(stderr, "type %s has multiple possible depths\n", hwloc_obj_type_string(type));
       return -1;
     } else if (depth == HWLOC_TYPE_DEPTH_UNKNOWN) {
-      fprintf(stderr, "type %s isn't available\n", hwloc_obj_type_string(type));
+      if (verbose >= 0)
+	fprintf(stderr, "type %s isn't available\n", hwloc_obj_type_string(type));
       return -1;
     }
     realtype = hwloc_get_depth_type(topology, depth);
-    if (type != realtype && verbose)
+    if (type != realtype && verbose > 0)
       fprintf(stderr, "using type %s (depth %d) instead of %s\n",
 	      hwloc_obj_type_string(realtype), depth, hwloc_obj_type_string(type));
     return depth;
@@ -178,8 +193,9 @@ hwloc_calc_depth_of_type(hwloc_topology_t topology, hwloc_obj_type_t type,
       for(i=0; ; i++) {
 	hwloc_obj_t obj = hwloc_get_obj_by_depth(topology, i, 0);
 	if (!obj) {
-	  fprintf(stderr, "Group with custom depth %d does not exist\n",
-		  depthattr);
+	  if (verbose >= 0)
+	    fprintf(stderr, "Group with custom depth %d does not exist\n",
+		    depthattr);
 	  return -1;
 	}
 	if (obj->type == type
@@ -188,10 +204,12 @@ hwloc_calc_depth_of_type(hwloc_topology_t topology, hwloc_obj_type_t type,
       }
     else if (type == HWLOC_OBJ_CACHE) {
       depth = hwloc_get_cache_type_depth(topology, depthattr, cachetype);
-      if (depth == HWLOC_TYPE_DEPTH_UNKNOWN)
-	fprintf(stderr, "Cache with custom depth %d and type %d does not exist\n", depthattr, cachetype);
-      else if (depth == HWLOC_TYPE_DEPTH_MULTIPLE)
-	fprintf(stderr, "Cache with custom depth %d and type %d has multiple possible depths\n", depthattr, cachetype);
+      if (verbose >= 0) {
+	if (depth == HWLOC_TYPE_DEPTH_UNKNOWN)
+	  fprintf(stderr, "Cache with custom depth %d and type %d does not exist\n", depthattr, cachetype);
+	else if (depth == HWLOC_TYPE_DEPTH_MULTIPLE)
+	  fprintf(stderr, "Cache with custom depth %d and type %d has multiple possible depths\n", depthattr, cachetype);
+      }
       return depth;
     } else
       assert(0);
@@ -216,7 +234,8 @@ hwloc_calc_parse_depth_prefix(hwloc_topology_t topology, unsigned topodepth,
   int err;
 
   if (typelen >= sizeof(typestring)) {
-    fprintf(stderr, "invalid type name %s\n", string);
+    if (verbose >= 0)
+      fprintf(stderr, "invalid type name %s\n", string);
     return -1;
   }
   strncpy(typestring, string, typelen);
@@ -232,11 +251,13 @@ hwloc_calc_parse_depth_prefix(hwloc_topology_t topology, unsigned topodepth,
   /* try to match a numeric depth */
   depth = strtol(string, &end, 0);
   if (end != &string[typelen]) {
-    fprintf(stderr, "invalid type name %s\n", string);
+    if (verbose >= 0)
+      fprintf(stderr, "invalid type name %s\n", string);
     return -1;
   }
   if ((unsigned) depth >= topodepth) {
-    fprintf(stderr, "ignoring invalid depth %u\n", depth);
+    if (verbose >= 0)
+      fprintf(stderr, "ignoring invalid depth %u\n", depth);
     return -1;
   }
   *typep = (hwloc_obj_type_t) -1;
@@ -244,15 +265,29 @@ hwloc_calc_parse_depth_prefix(hwloc_topology_t topology, unsigned topodepth,
 }
 
 static __hwloc_inline int
-hwloc_calc_parse_range(const char *string,
+hwloc_calc_parse_range(const char *_string,
 		       int *firstp, int *amountp, int *stepp, int *wrapp,
 		       const char **dotp)
 {
-  const char *dash, *dot, *colon;
-  int first, amount, step, wrap;
+  char string[65];
+  size_t len;
+  char *dot, *end, *end2;
+  long first, last, amount;
+  int wrap;
 
-  dot = strchr(string, '.');
+  dot = strchr(_string, '.');
   *dotp = dot;
+  if (dot) {
+    len = dot - _string;
+  } else {
+    len = strlen(_string);
+  }
+  if (len >= sizeof(string)) {
+    fprintf(stderr, "invalid range `%s', too long\n", string);
+    return -1;
+  }
+  memcpy(string, _string, len);
+  string[len] = '\0';
 
   if (!isdigit(*string)) {
     if (!strncmp(string, "all", 3)) {
@@ -273,32 +308,49 @@ hwloc_calc_parse_range(const char *string,
       *stepp = 2;
       *wrapp = 0;
       return 0;
-    } else
+    } else {
+      fprintf(stderr, "unrecognized range keyword `%s'\n", string);
       return -1;
+    }
   }
 
-  first = atoi(string);
+  first = strtol(string, &end, 10);
   amount = 1;
-  step = 1;
   wrap = 0;
 
-  dash = strchr(string, '-');
-  if (dash && (dash < dot || !dot)) {
-    if (*(dash+1) == '\0')
+  if (*end == '-') {
+    last = strtol(end+1, &end2, 10);
+    if (*end2) {
+      fprintf(stderr, "invalid character at `%s' after range at `%s'\n", end2, string);
+      return -1;
+    } else if (end2 == end+1) {
+      /* X- */
       amount = -1;
-    else
-      amount = atoi(dash+1)-first+1;
-  } else {
-    colon = strchr(string, ':');
-    if (colon && (colon < dot || !dot)) {
-      amount = atoi(colon+1);
-      wrap = 1;
+    } else {
+      /* X-Y */
+      amount = last-first+1;
     }
+
+  } else if (*end == ':') {
+    /* X:Y */
+    wrap = 1;
+    amount = strtol(end+1, &end2, 10);
+    if (*end2) {
+      fprintf(stderr, "invalid character at `%s' after range at `%s'\n", end2, string);
+      return -1;
+    } else if (end2 == end+1) {
+      fprintf(stderr, "missing width at `%s' in range at `%s'\n", end2, string);
+      return -1;
+    }
+
+  } else if (*end) {
+    fprintf(stderr, "invalid character at `%s' after index at `%s'\n", end, string);
+    return -1;
   }
 
   *firstp = first;
   *amountp = amount;
-  *stepp = step;
+  *stepp = 1;
   *wrapp = wrap;
   return 0;
 }
@@ -307,7 +359,9 @@ static __hwloc_inline int
 hwloc_calc_append_object_range(hwloc_topology_t topology, unsigned topodepth,
 			       hwloc_const_bitmap_t rootset, int depth,
 			       const char *string, /* starts with indexes following the colon */
-			       int logical, hwloc_bitmap_t set, int verbose)
+			       int logical,
+			       void (*cbfunc)(void *, hwloc_obj_t, int), void *cbdata,
+			       int verbose)
 {
   hwloc_obj_t obj;
   unsigned width;
@@ -351,7 +405,7 @@ hwloc_calc_append_object_range(hwloc_topology_t topology, unsigned topodepth,
       i = 0;
 
     obj = hwloc_calc_get_obj_inside_cpuset_by_depth(topology, rootset, depth, i, logical);
-    if (verbose || !obj) {
+    if (verbose > 0 || (!obj && verbose >= 0)) {
       char *s;
       hwloc_bitmap_asprintf(&s, rootset);
       if (obj)
@@ -364,12 +418,12 @@ hwloc_calc_append_object_range(hwloc_topology_t topology, unsigned topodepth,
     }
     if (obj) {
       if (dot) {
-	hwloc_calc_append_object_range(topology, topodepth, obj->cpuset, nextdepth, nextsep+1, logical, set, verbose);
+	hwloc_calc_append_object_range(topology, topodepth, obj->cpuset, nextdepth, nextsep+1, logical, cbfunc, cbdata, verbose);
       } else {
 	/* add to the temporary cpuset
 	 * and let the caller add/clear/and/xor for the actual final cpuset depending on cmdline options
 	 */
-        hwloc_calc_append_cpuset(set, obj->cpuset, HWLOC_CALC_APPEND_ADD, verbose);
+        cbfunc(cbdata, obj, verbose);
       }
     }
   }
@@ -378,15 +432,11 @@ hwloc_calc_append_object_range(hwloc_topology_t topology, unsigned topodepth,
 }
 
 static __hwloc_inline int
-hwloc_calc_append_iodev(hwloc_bitmap_t set, hwloc_obj_t obj,
-			hwloc_calc_append_mode_t mode, int verbose)
+hwloc_calc_append_iodev(void (*cbfunc)(void *, hwloc_obj_t, int), void *cbdata,
+			hwloc_obj_t obj,
+			int verbose)
 {
-  while (obj && !obj->cpuset)
-    obj = obj->parent;
-  if (!obj)
-    /* do nothing */
-    return 0;
-  hwloc_calc_append_cpuset(set, obj->cpuset, mode, verbose);
+  cbfunc(cbdata, obj, verbose);
   return 0;
 }
 
@@ -417,7 +467,9 @@ hwloc_calc_find_next_pci_object(hwloc_topology_t topology, int vendor, int devic
 }
 
 static __hwloc_inline int
-hwloc_calc_append_pci_object_range(hwloc_topology_t topology, const char *string, hwloc_bitmap_t set, int verbose)
+hwloc_calc_append_pci_object_range(hwloc_topology_t topology, const char *string,
+				   void (*cbfunc)(void *, hwloc_obj_t, int), void *cbdata,
+				   int verbose)
 {
   hwloc_obj_t obj;
   int vendor, device;
@@ -461,13 +513,14 @@ hwloc_calc_append_pci_object_range(hwloc_topology_t topology, const char *string
   for(oldi=-1, i=first, j=0; j<amount || amount == -1; oldi=i, i+=step, j++) {
     obj = hwloc_calc_find_next_pci_object(topology, vendor, device, obj, i-oldi, wrap);
     if (obj) {
-      if (verbose)
+      if (verbose > 0)
 	printf("using matching PCI object #%d bus id %04x:%02x:%02x.%01x\n", i,
 	       obj->attr->pcidev.domain, obj->attr->pcidev.bus, obj->attr->pcidev.dev, obj->attr->pcidev.func);
-      hwloc_calc_append_iodev(set, obj, HWLOC_CALC_APPEND_ADD, verbose);
+      hwloc_calc_append_iodev(cbfunc, cbdata, obj, verbose);
     } else {
       if (amount != -1)
-	fprintf(stderr, "no matching PCI object #%d\n", i);
+	if (verbose >= 0)
+	  fprintf(stderr, "no matching PCI object #%d\n", i);
       break;
     }
   }
@@ -478,10 +531,72 @@ hwloc_calc_append_pci_object_range(hwloc_topology_t topology, const char *string
    * but we don't want some ugly and unmaintainable code
    */
 
-  fprintf(stderr, "invalid PCI device %s\n", string);
+  if (verbose >= 0)
+    fprintf(stderr, "invalid PCI device %s\n", string);
   return -1;
 }
 
+static __hwloc_inline int
+hwloc_calc_process_type_arg(hwloc_topology_t topology, unsigned topodepth,
+			    const char *arg, size_t typelen,
+			    int logical,
+			    void (*cbfunc)(void *, hwloc_obj_t, int), void *cbdata,
+			    int verbose)
+{
+  const char *sep = &arg[typelen];
+  hwloc_obj_type_t type = (hwloc_obj_type_t) -1;
+  int depth;
+
+  depth = hwloc_calc_parse_depth_prefix(topology, topodepth,
+					arg, typelen,
+					&type,
+					verbose);
+  if (depth < 0) {
+    /* if we didn't find a depth but found a type, handle special cases */
+    hwloc_obj_t obj = NULL;
+    if (*sep == ':' && type == HWLOC_OBJ_PCI_DEVICE) {
+      /* FIXME: change to another syntax? */
+      return hwloc_calc_append_pci_object_range(topology, sep+1, cbfunc, cbdata, verbose);
+
+    } else if (*sep == '=' && type == HWLOC_OBJ_PCI_DEVICE) {
+      /* try to match a busid */
+      obj = hwloc_get_pcidev_by_busidstring(topology, sep+1);
+      if (obj)
+	return hwloc_calc_append_iodev(cbfunc, cbdata, obj, verbose);
+      if (verbose >= 0)
+	fprintf(stderr, "invalid PCI device %s\n", sep+1);
+      return -1;
+
+    } else if (*sep == '=' && type == HWLOC_OBJ_OS_DEVICE) {
+      /* try to match a OS device name */
+      while ((obj = hwloc_get_next_osdev(topology, obj)) != NULL) {
+	if (!strcmp(obj->name, sep+1))
+	  return hwloc_calc_append_iodev(cbfunc, cbdata, obj, verbose);
+      }
+      if (verbose >= 0)
+	fprintf(stderr, "invalid OS device %s\n", sep+1);
+      return -1;
+    } else
+      return -1;
+  }
+
+  /* look at indexes following this type/depth */
+  return hwloc_calc_append_object_range(topology, topodepth, hwloc_topology_get_complete_cpuset(topology), depth, sep+1, logical, cbfunc, cbdata, verbose);
+}
+
+static __hwloc_inline void
+hwloc_calc_process_arg_cpuset_cb(void *_data, hwloc_obj_t obj, int verbose)
+{
+  hwloc_bitmap_t set = _data;
+  /* walk up out of I/O objects */
+  while (obj && !obj->cpuset)
+    obj = obj->parent;
+  if (!obj)
+    /* do nothing */
+    return;
+  hwloc_calc_append_cpuset(set, obj->cpuset, HWLOC_CALC_APPEND_ADD, verbose);
+}
+
 static __hwloc_inline int
 hwloc_calc_process_arg(hwloc_topology_t topology, unsigned topodepth,
 		       const char *arg, int logical, hwloc_bitmap_t set,
@@ -508,50 +623,11 @@ hwloc_calc_process_arg(hwloc_topology_t topology, unsigned topodepth,
   /* try to match a type/depth followed by a special character */
   typelen = strspn(arg, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
   if (typelen && (arg[typelen] == ':' || arg[typelen] == '=')) {
-    const char *sep = &arg[typelen];
-    hwloc_obj_type_t type = (hwloc_obj_type_t) -1;
-    int depth;
-    hwloc_bitmap_t newset;
-
-    depth = hwloc_calc_parse_depth_prefix(topology, topodepth,
-					  arg, typelen,
-					  &type,
-					  verbose);
-    if (depth < 0) {
-      /* if we didn't find a depth but found a type, handle special cases */
-      hwloc_obj_t obj = NULL;
-      if (*sep == ':' && type == HWLOC_OBJ_PCI_DEVICE) {
-	/* FIXME: change to another syntax? */
-	/* agregate the whole range in newset before applying it to set */
-	newset = hwloc_bitmap_alloc();
-	err = hwloc_calc_append_pci_object_range(topology, sep+1, newset, verbose);
-	if (!err)
-	  err = hwloc_calc_append_cpuset(set, newset, mode, verbose);
-	hwloc_bitmap_free(newset);
-	return err;
-      } else if (*sep == '=' && type == HWLOC_OBJ_PCI_DEVICE) {
-	/* try to match a busid */
-	obj = hwloc_get_pcidev_by_busidstring(topology, sep+1);
-	if (obj)
-	  return hwloc_calc_append_iodev(set, obj, mode, verbose);
-	fprintf(stderr, "invalid PCI device %s\n", sep+1);
-	return -1;
-      } else if (*sep == '=' && type == HWLOC_OBJ_OS_DEVICE) {
-	/* try to match a OS device name */
-	while ((obj = hwloc_get_next_osdev(topology, obj)) != NULL) {
-	  if (!strcmp(obj->name, sep+1))
-	    return hwloc_calc_append_iodev(set, obj, mode, verbose);
-	}
-	fprintf(stderr, "invalid OS device %s\n", sep+1);
-	return -1;
-      } else
-	return -1;
-    }
-
-    /* look at indexes following this type/depth */
-    /* agregate the whole range in newset before applying it to set */
-    newset = hwloc_bitmap_alloc();
-    err = hwloc_calc_append_object_range(topology, topodepth, hwloc_topology_get_complete_cpuset(topology), depth, sep+1, logical, newset, verbose);
+    /* process type/depth */
+    hwloc_bitmap_t newset = hwloc_bitmap_alloc();
+    err = hwloc_calc_process_type_arg(topology, topodepth, arg, typelen, logical,
+				      hwloc_calc_process_arg_cpuset_cb, newset,
+				      verbose);
     if (!err)
       err = hwloc_calc_append_cpuset(set, newset, mode, verbose);
     hwloc_bitmap_free(newset);
